from rdflib.query import Processor, Result, UpdateProcessor
from rdflib.plugins.sparql.sparql import Query
from rdflib.plugins.sparql.parser import parseQuery, parseUpdate

from quit.tools.algebra import translateQuery, translateUpdate
from quit.tools.evaluate import evalQuery
from quit.tools.update import evalUpdate

class SPARQLUpdateProcessor(UpdateProcessor):
    def __init__(self, graph):
        self.graph = graph

    def update(self, strOrQuery, initBindings={}, initNs={}):
        if isinstance(strOrQuery, str):
            strOrQuery=translateUpdate(parseUpdate(strOrQuery), initNs=initNs)

        return evalUpdate(self.graph, strOrQuery, initBindings)


class SPARQLProcessor(Processor):

    def __init__(self, graph):
        self.graph = graph

    def query(
            self, strOrQuery, initBindings={},
            initNs={}, base=None, DEBUG=False):
        """
        Evaluate a query with the given initial bindings, and initial
        namespaces. The given base is used to resolve relative URIs in
        the query and will be overridden by any BASE given in the query.
        """

        if not isinstance(strOrQuery, Query):
            parsetree = parseQuery(strOrQuery)
            query = translateQuery(parsetree, base, initNs)
        else:
            query = strOrQuery

        return evalQuery(self.graph, query, initBindings, base)


from flask import Blueprint
from flask import session, request, current_app, url_for
import logging
import urllib
import uuid
import json

logger = logging.getLogger('quit.modules.application')

application = Blueprint('application', __name__)


@application.route("/login", methods=['GET', 'POST'])
def login():
    if "state" not in session:
        state = str(uuid.uuid4())
        session["state"] = state
    else:
        state = session["state"]
    logger.debug("request url: {}".format(request.url))
    redirect_uri = request.url
    authorizeEndpoint = "https://github.com/login/oauth/authorize"
    tokenEndpoint = "https://github.com/login/oauth/access_token"

    error = request.values.get('error', None)
    code = request.values.get('code', None)

    config = current_app.config['quit'].config

    if error or code is None:
        params = {'client_id': config.oauthclientid,
                  'redirect_uri': redirect_uri,
                  'scope': 'repo',
                  'state': state}
        loginURL = authorizeEndpoint + "?" + urllib.parse.urlencode(params)
        return "<a href='{}'>Login with GitHub</a>".format(loginURL)
    else:
        request_state = request.values.get('state', None)
        if not state == request_state:
            return "Error"
        post_data = {'client_id': config.oauthclientid,
                     'client_secret': config.oauthclientsecret,
                     'code': code,
                     'state': state}
        tokenrequest = urllib.request.Request(tokenEndpoint,
                                              urllib.parse.urlencode(post_data).encode())
        tokenrequest.add_header('Accept', 'application/json')
        tokenresponse = urllib.request.urlopen(tokenrequest).read().decode("utf-8")
        token = json.loads(tokenresponse)["access_token"]
        session["OAUTH_TOKEN"] = token
        return "success (<a href='{}'>back to quit</a>)".format(url_for("git.commits"))


@application.route("/logout", methods=['GET', 'POST'])
def logout():
    print(session)
    print("logout")
    session.clear()
    print("logout")
    print(session)
    return "you are successfully logged out (<a href='{}'>back to quit</a>)".format(
           url_for("git.commits"))


def isLoggedIn():
    """Returns true if OAUTH_TOKEN is set in the session."""
    if "OAUTH_TOKEN" in session:
        return True
    return False


def githubEnabled():
    """Returns true if oauthclientid and oauthclientsecret are configured."""
    config = current_app.config['quit'].config

    if config.oauthclientid and config.oauthclientsecret:
        return True
    return False



from flask import Flask
from flask_uwsgi_websocket import GeventWebSocket

app = Flask(__name__.split('.')[0])

print("name is: {}".format(__name__.split('.')[0]))
print(app.name)
print(app.import_name)

websocket = GeventWebSocket(app)

@websocket.route('/echo')
def echo(ws):
    while True:
        msg = ws.receive()
        ws.send(msg)

@app.route('/hello')
def hallo():
    return 'Hello'

if __name__ == '__main__':
    app.run(gevent=100)



from __future__ import with_statement
from flask import Response
import os
import contextlib
import signal
import sys
from datetime import tzinfo, timedelta, datetime
from quit.graphs import InMemoryAggregatedGraph
from collections import OrderedDict
from urllib.parse import quote_plus, urlparse


ZERO = timedelta(0)
HOUR = timedelta(hours=1)


class TZ(tzinfo):
    """Fixed offset in minutes east from UTC."""

    def __init__(self, offset, name):
        self.__offset = timedelta(minutes=offset)
        self.__name = name

    def utcoffset(self, dt):
        return self.__offset

    def tzname(self, dt):
        return self.__name

    def dst(self, dt):
        return ZERO


def git_timestamp(ts, offset):
    import quit.utils as tzinfo
    if offset == 0:
        tz = tzinfo.TZ(0, "UTC")
    else:
        hours, rem = divmod(abs(offset), 60)
        tzname = 'UTC%+03d:%02d' % ((hours, -hours)[offset < 0], rem)
        tz = tzinfo.TZ(offset, tzname)
    return datetime.fromtimestamp(ts, tz)


def iri_to_name(iri):
    parsedIri = urlparse(iri)
    nameParts = [parsedIri.netloc]
    if parsedIri.path.strip("/"):
        nameParts += parsedIri.path.strip("/").split("/")
    return quote_plus("_".join(nameParts))


def graphdiff(first, second):
    """
    Diff between graph instances, should be replaced/included in quit diff
    """
    from rdflib.compare import to_isomorphic, graph_diff

    diffs = OrderedDict()
    iris = set()

    if first is not None and isinstance(first, InMemoryAggregatedGraph):
        first_identifiers = list((g.identifier for g in first.graphs()))
        iris = iris.union(first_identifiers)
    if second is not None and isinstance(second, InMemoryAggregatedGraph):
        second_identifiers = list((g.identifier for g in second.graphs()))
        iris = iris.union(second_identifiers)

    for iri in sorted(list(iris)):
        changes = diffs.get(iri, [])

        if (
            first is not None and iri in first_identifiers
        ) and (
            second is not None and iri in second_identifiers
        ):
            g1 = first.get_context(iri)
            g2 = second.get_context(iri)
            in_both, in_first, in_second = graph_diff(to_isomorphic(g1), to_isomorphic(g2))

            if len(in_second) > 0:
                changes.append(('additions', ((s, p, o) for s, p, o in in_second)))
            if len(in_first) > 0:
                changes.append(('removals', ((s, p, o) for s, p, o in in_first)))
        elif first is not None and iri in first_identifiers:
            changes.append(('removals', ((s, p, o) for s, p, o in first.get_context(iri))))
        elif second is not None and iri in second_identifiers:
            changes.append(('additions', ((s, p, o) for s, p, o in second.get_context(iri))))
        else:
            continue

        diffs[iri] = changes
    return diffs


def _sigterm_handler(signum, frame):
    sys.exit(0)


_sigterm_handler.__enter_ctx__ = False


@contextlib.contextmanager
def handle_exit(callback=None, append=False):
    """A context manager which properly handles SIGTERM and SIGINT
    (KeyboardInterrupt) signals, registering a function which is
    guaranteed to be called after signals are received.
    Also, it makes sure to execute previously registered signal
    handlers as well (if any).

    >>> app = App()
    >>> with handle_exit(app.stop):
    ...     app.start()
    ...
    >>>

    If append == False raise RuntimeError if there's already a handler
    registered for SIGTERM, otherwise both new and old handlers are
    executed in this order.
    """
    old_handler = signal.signal(signal.SIGTERM, _sigterm_handler)
    if (old_handler != signal.SIG_DFL) and (old_handler != _sigterm_handler):
        if not append:
            raise RuntimeError("there is already a handler registered for "
                               "SIGTERM: %r" % old_handler)

        def handler(signum, frame):
            try:
                _sigterm_handler(signum, frame)
            finally:
                old_handler(signum, frame)
        signal.signal(signal.SIGTERM, handler)

    if _sigterm_handler.__enter_ctx__:
        raise RuntimeError("can't use nested contexts")
    _sigterm_handler.__enter_ctx__ = True

    try:
        yield
    except KeyboardInterrupt:
        pass
    except SystemExit as err:
        # code != 0 refers to an application error (e.g. explicit
        # sys.exit('some error') call).
        # We don't want that to pass silently.
        # Nevertheless, the 'finally' clause below will always
        # be executed.
        if err.code != 0:
            raise
    finally:
        _sigterm_handler.__enter_ctx__ = False
        if callback is not None:
            callback()


if __name__ == '__main__':
    # ===============================================================
    # --- test suite
    # ===============================================================

    import unittest
    import os

    class TestOnExit(unittest.TestCase):

        def setUp(self):
            # reset signal handlers
            signal.signal(signal.SIGTERM, signal.SIG_DFL)
            self.flag = None

        def tearDown(self):
            # make sure we exited the ctx manager
            self.assertTrue(self.flag is not None)

        def test_base(self):
            with handle_exit():
                pass
            self.flag = True

        def test_callback(self):
            callback = []
            with handle_exit(lambda: callback.append(None)):
                pass
            self.flag = True
            self.assertEqual(callback, [None])

        def test_kinterrupt(self):
            with handle_exit():
                raise KeyboardInterrupt
            self.flag = True

        def test_sigterm(self):
            with handle_exit():
                os.kill(os.getpid(), signal.SIGTERM)
            self.flag = True

        def test_sigint(self):
            with handle_exit():
                os.kill(os.getpid(), signal.SIGINT)
            self.flag = True

        def test_sigterm_old(self):
            # make sure the old handler gets executed
            queue = []
            signal.signal(signal.SIGTERM, lambda s, f: queue.append('old'))
            with handle_exit(lambda: queue.append('new'), append=True):
                os.kill(os.getpid(), signal.SIGTERM)
            self.flag = True
            self.assertEqual(queue, ['old', 'new'])

        def test_sigint_old(self):
            # make sure the old handler gets executed
            queue = []
            signal.signal(signal.SIGINT, lambda s, f: queue.append('old'))
            with handle_exit(lambda: queue.append('new'), append=True):
                os.kill(os.getpid(), signal.SIGINT)
            self.flag = True
            self.assertEqual(queue, ['old', 'new'])

        def test_no_append(self):
            # make sure we can't use the context manager if there's
            # already a handler registered for SIGTERM
            signal.signal(signal.SIGTERM, lambda s, f: sys.exit(0))
            try:
                with handle_exit(lambda: self.flag.append(None)):
                    pass
            except RuntimeError:
                pass
            else:
                self.fail("exception not raised")
            finally:
                self.flag = True

        def test_nested_context(self):
            self.flag = True
            try:
                with handle_exit():
                    with handle_exit():
                        pass
            except RuntimeError:
                pass
            else:
                self.fail("exception not raised")

    unittest.main()



#!/usr/bin/env python3


class Error(Exception):
    pass


class InvalidConfigurationError(Error):
    pass


class MissingConfigurationError(Error):
    pass


class MissingFileError(Error):
    pass


class RepositoryNotFound(Error):
    """
    Exception raised when a repository is invalid
    """


class ResourceNotFound(Error):
    """
    Thrown when a non-existent resource is requested
    """


class RevisionNotFound(ResourceNotFound):
    """
    Thrown when a non-existent revision is requested
    """

    def __init__(self, id):
        ResourceNotFound.__init__(self, "No commit '%s' in the repository," % id)


class NodeNotFound(ResourceNotFound):
    """
    Thrown when a non-existent node is requested
    """

    def __init__(self, path, id):
        ResourceNotFound.__init__(self, "No node '%s' in commit '%s'," % (path, id))


class IndexError(Error):
    """
    Thrown during indexing
    """
    pass


class ServiceException(Error):
    """
    Thrown when requesting a missing service
    """
    pass


class UnSupportedQuery(Exception):
    """
    Thrown when providing a query which includes an unsupported keyword
    """

    def __init__(self, message=None):
        self.message = message
        pass

    def __str__(self):
        if self.message is not None:
            return ("This query is not supported by this endpoint: {}".format(self.message))
        else:
            return ("This query is not supported by this endpoint")


class UnSupportedQueryType(UnSupportedQuery):
    """
    Thrown when providing an unsupported query type
    """
    pass


class SparqlProtocolError(Error):
    pass


class FromNamedError(Error):
    pass


class NonAbsoluteBaseError(Error):
    """
    Thrown when a query/update contains a non absolute BASE URI
    """
    pass


class UnknownConfigurationError(Error):
    pass


class QuitGitRepoError(Error):
    pass


class RemoteNotFound(QuitGitRepoError):
    """Raised when a requested remote is not configured on the repository."""
    pass


class QuitGitRefNotFound(QuitGitRepoError):
    """Raised when a reference could not be found."""
    pass


class QuitGitPushError(QuitGitRepoError):
    """Raised when it is not possible to push to a remote repository."""
    pass


class QuitMergeConflict(QuitGitRepoError):
    """Raised for a merge conflict."""

    def __init__(self, message=None, blobConflicts=None):
        self.message = message
        self.blobConflicts = blobConflicts

    def getObject(self):
        return self.blobConflicts


class QuitBlobMergeConflict(QuitGitRepoError):
    """Raised for a merge conflict."""

    def __init__(self, message=None, ok=None, conflicts=None):
        self.message = message
        self.ok = ok
        self.conflicts = conflicts

    def getObject(self):
        return {
            "message": self.message,
            "ok": "\n".join(self.ok),
            "conflicts": self.conflicts
        }



import traceback
import pygit2

from flask import Blueprint, request, current_app, make_response
from werkzeug.http import parse_accept_header
from quit.web.app import render_template
from quit.web.extras.commits_graph import CommitGraph, generate_graph_data
from quit.exceptions import QuitMergeConflict
from quit.utils import git_timestamp
from quit.web.modules.application import isLoggedIn, githubEnabled
import json
import logging
import re

logger = logging.getLogger('quit.web.modules.git')
__all__ = ('git')

git = Blueprint('git', __name__)


@git.route("/commits", defaults={'branch_or_ref': None}, methods=['GET'])
@git.route("/commits/<path:branch_or_ref>", methods=['GET'])
def commits(branch_or_ref):
    """
    Lists all commits of a given git branch.

    Returns:
    HTTP Response 200: a list of commits
    HTTP Response 403: unknown branch or ref
    HTTP Response 406: Unsupported Mimetype requested
    """
    quit = current_app.config['quit']

    if not branch_or_ref:
        branch_or_ref = quit.getDefaultBranch()

    try:
        current_app.logger.debug(branch_or_ref)
        if not quit.repository.is_empty:
            results = quit.repository.revisions(branch_or_ref, order=pygit2.GIT_SORT_TIME)
        else:
            results = []

        if 'Accept' in request.headers:
            mimetype = parse_accept_header(request.headers['Accept']).best
        else:
            mimetype = '*/*'

        if mimetype in ['text/html', 'application/xhtml_xml', '*/*']:
            data = generate_graph_data(CommitGraph.gets(results))
            response = make_response(render_template('commits.html', results=results, data=data,
                                                     current_ref=branch_or_ref,
                                                     isLoggedIn=isLoggedIn,
                                                     githubEnabled=githubEnabled))
            response.headers['Content-Type'] = 'text/html'
            return response
        elif mimetype in ['application/json', 'application/sparql-results+json']:
            res = []
            for revision in results:
                res.append({"id": revision.id,
                            "author_name": revision.author.name,
                            "author_email": revision.author.email,
                            "author_time": str(git_timestamp(revision.author.time,
                                                             revision.author.offset)),
                            "committer_name": revision.committer.name,
                            "committer_email": revision.committer.email,
                            "committer_time": str(git_timestamp(revision.committer.time,
                                                                revision.committer.offset)),
                            "committer_offset": revision.committer.offset,
                            "message": revision.message,
                            "parrents": [parent.id for parent in revision.parents]})
            response = make_response(json.dumps(res), 200)
            response.headers['Content-Type'] = 'application/json'
            return response
        else:
            return "<pre>Unsupported Mimetype: {}</pre>".format(mimetype), 406
    except Exception as e:
        current_app.logger.error(e)
        current_app.logger.error(traceback.format_exc())
        return "<pre>" + traceback.format_exc() + "</pre>", 403


@git.route("/pull", defaults={'remote': None, "refspec": None}, methods=['POST', 'GET'])
@git.route("/pull/<remote>", defaults={"refspec": None}, methods=['GET', 'POST'])
@git.route("/pull/<remote>/<path:refspec>", methods=['GET', 'POST'])
def pull(remote, refspec):
    """Pull from remote.

    Arguments:
    remote -- The remote repository that is the source of the pull operation. The remote with
    its url has to be configured on the repository.
    refspec -- Specifies which refs to fetch and which local refs to update.

    Returns:
    HTTP Response 200: If pull was possible
    HTTP Response 201: If pull was possible and new commits were fetched or a merge commit was
                       created*
    HTTP Response 400: If pull did not work
    HTTP Response 409: If pull produces a conflict*
    (* not yet implemented)
    """
    if remote is None:
        remote = request.values.get('remote', None)
    if refspec is None:
        refspec = request.values.get('refspec', None)

    try:
        current_app.config['quit'].repository.pull(remote_name=remote, refspec=refspec)
        current_app.config['quit'].syncAll()
        return '', 200
    except Exception as e:
        current_app.logger.error(e)
        current_app.logger.error(traceback.format_exc())
        return "<pre>" + traceback.format_exc() + "</pre>", 400


@git.route("/fetch", defaults={'remote': None, "refspec": None}, methods=['POST', 'GET'])
@git.route("/fetch/<remote>", defaults={"refspec": None}, methods=['GET', 'POST'])
@git.route("/fetch/<remote>/<path:refspec>", methods=['GET', 'POST'])
def fetch(remote, refspec):
    """Fetch from remote.

    Arguments:
    remote -- The remote repository that is the source of the fetch operation. The remote with its
              url has to be configured on the repository.
    refspec -- Specifies which refs to fetch and which local refs to update.

    Returns:
    HTTP Response 200: If fetch was possible
    HTTP Response 201: If fetch was possible and new commits were fetched*
    HTTP Response 400: If fetch did not work
    HTTP Response 409: If fetch produces a conflict*
    (* not yet implemented)
    """
    if remote is None:
        remote = request.values.get('remote', None)
    if refspec is None:
        refspec = request.values.get('refspec', None)

    try:
        current_app.config['quit'].repository.fetch(remote, refspec)
        current_app.config['quit'].syncAll()
        return '', 200
    except Exception as e:
        current_app.logger.error(e)
        current_app.logger.error(traceback.format_exc())
        return "<pre>" + traceback.format_exc() + "</pre>", 400


@git.route("/push", defaults={'remote': None, "refspec": None}, methods=['POST', 'GET'])
@git.route("/push/<remote>", defaults={"refspec": None}, methods=['GET', 'POST'])
@git.route("/push/<remote>/<path:refspec>", methods=['GET', 'POST'])
def push(remote, refspec):
    """Push to remote.

    Arguments:
    remote -- The remote repository that is the destination of the push operation. The remote with
              its url has to be configured on the repository.
    refspec -- Specifies which refs to push and which remote refs to update.

    Returns:
    HTTP Response 200: If push was possible
    HTTP Response 400: If push did not work
    HTTP Response 409: If push produces a conflict on the remote end*
    (* not yet implemented)
    """
    if remote is None:
        remote = request.values.get('remote', None)
    if refspec is None:
        refspec = request.values.get('refspec', None)

    try:
        current_app.config['quit'].repository.push(remote, refspec)
        return '', 200
    except Exception as e:
        current_app.logger.error(e)
        current_app.logger.error(traceback.format_exc())
        return "<pre>" + traceback.format_exc() + "</pre>", 400


@git.route("/merge", defaults={'refspec': None}, methods=['GET', 'POST'])
@git.route("/merge/<path:refspec>", methods=['GET', 'POST'])
def merge(refspec):
    """Merge branch into target (refspec=<branch>:<target>).

    Merge 'branch' into 'target' and set 'target' to the resulting commit.

    Returns:
    HTTP Response 200: If merge was possible
    HTTP Response 201: If merge was possible and a merge commit was created
    HTTP Response 202: If merge was possible and a fast-forward happened
    HTTP Response 400: If merge did not work
    HTTP Response 409: If merge produces a conflict
    """
    quit = current_app.config['quit']

    try:
        if 'Accept' in request.headers:
            mimetype = parse_accept_header(request.headers['Accept']).best
        else:
            mimetype = '*/*'

        if mimetype in ['text/html', 'application/xhtml_xml']:
            response = make_response(render_template('merge.html'))
            response.headers['Content-Type'] = 'text/html'
            return response
        elif mimetype in ['application/json', '*/*']:
            # Actual Merge
            if refspec is None:
                refspec = request.values.get('refspec', None)

            if refspec:
                try:
                    branch, target = refspec.split(":")
                except ValueError:
                    branch = refspec
                    target = None
            else:
                branch = request.values.get('branch', None)
                target = request.values.get('target', None)
            method = request.values.get('method', None)
            try:
                result = quit.repository.merge(target=target, branch=branch, method=method)
            except QuitMergeConflict as mergeconflict:
                response = make_response(json.dumps(mergeconflict.getObject()), 409)
                response.headers['Content-Type'] = 'application/json'
                return response

            resultMessage = ""
            resultCode = 200

            if (result & pygit2.GIT_MERGE_ANALYSIS_FASTFORWARD and result &
                    pygit2.GIT_MERGE_ANALYSIS_UNBORN):
                resultMessage = "{target} was unborn and is now set to {branch}".format(
                    target=target, branch=branch)
                resultCode = 202
            elif result & pygit2.GIT_MERGE_ANALYSIS_FASTFORWARD:
                resultMessage = "{target} could be fast-forwarded to {branch}".format(
                    target=target, branch=branch)
                resultCode = 202
            elif result & pygit2.GIT_MERGE_ANALYSIS_UP_TO_DATE:
                resultMessage = "{target} is already up-to-date with {branch}".format(
                    target=target, branch=branch)
                resultCode = 200
            elif result & pygit2.GIT_MERGE_ANALYSIS_NORMAL:
                resultMessage = "{branch} was merged into {target}, merge commit created".format(
                    target=target, branch=branch)
                resultCode = 201
            else:
                resultMessage = "don't know what happened to {target} and {branch}".format(
                    target=target, branch=branch)

            result_object = {"status": result, "result_message": resultMessage}

            quit.syncAll()
            response = make_response(json.dumps(result_object), resultCode)
            response.headers['Content-Type'] = 'application/json'
            return response
        else:
            return "<pre>Unsupported Mimetype: {}</pre>".format(mimetype), 406

    except Exception as e:
        logger.error(e)
        logger.error(traceback.format_exc())
        return "<pre>" + traceback.format_exc() + "</pre>", 400


@git.route("/branch", defaults={'refspec': None}, methods=['GET', 'POST'])
@git.route("/branch/<path:refspec>", methods=['GET', 'POST'])
def branch(refspec):
    """Branch two commits and set the result to branch_or_ref.

    merge branch into target and set branch_or_ref to the resulting commit
    - if only branch_or_ref is given, do nothing
    - if branch_or_ref and branch is given, merge branch into branch_or_ref and set branch_or_ref to
        the resulting commit
    - if branch_or_ref, branch and target are given, merge branch into target and set branch_or_ref
        to the resulting commit

    Returns:
    HTTP Response 200: If requesting the HTML interface
    HTTP Response 201: If branch was possible
    HTTP Response 400: If branching did not work or unsupported mimetype
    """
    quit = current_app.config['quit']

    try:
        if 'Accept' in request.headers:
            mimetype = parse_accept_header(request.headers['Accept']).best
        else:
            mimetype = '*/*'

        status = 200

        if refspec:
            oldbranch, newbranch = refspec.split(":")
        else:
            oldbranch = request.values.get('oldbranch')
            newbranch = request.values.get('newbranch')
        if newbranch:
            quit.repository.branch(oldbranch, newbranch)
            status = 201

        if mimetype in ['text/html', 'application/xhtml_xml', '*/*']:
            response = make_response(render_template('branch.html'))
            response.headers['Content-Type'] = 'text/html'
            return response, status
        else:
            return "<pre>Unsupported Mimetype: {}</pre>".format(mimetype), 400

    except Exception as e:
        current_app.logger.error(e)
        current_app.logger.error(traceback.format_exc())
        return "<pre>" + traceback.format_exc() + "</pre>", 400


@git.route("/delete/branch", defaults={'refspec': None}, methods=['GET', 'POST'])
@git.route("/delete/branch/<path:refspec>", methods=['GET', 'POST'])
def del_branch(refspec):
    """Branch two commits and set the result to branch_or_ref.

    merge branch into target and set branch_or_ref to the resulting commit
    - if only branch_or_ref is given, do nothing
    - if branch_or_ref and branch is given, merge branch into branch_or_ref and set branch_or_ref to
        the resulting commit
    - if branch_or_ref, branch and target are given, merge branch into target and set branch_or_ref
        to the resulting commit

    Returns:
    HTTP Response 200: If requesting the HTML interface
    HTTP Response 201: If branch was possible
    HTTP Response 400: If branching did not work or unsupported mimetype
    """
    quit = current_app.config['quit']

    try:
        if 'Accept' in request.headers:
            mimetype = parse_accept_header(request.headers['Accept']).best
        else:
            mimetype = '*/*'

        refspec = re.sub("^refs/heads/", "", refspec)

        branch = quit.repository._repository.branches.get(refspec)
        if branch:
            branch.delete()
            message = "{} deleted".format(refspec)
            status = 200
        else:
            message = "{} not found".format(refspec)
            status = 404

        if mimetype in ['text/html', 'application/xhtml_xml', '*/*']:
            response = make_response(render_template('message.html', message=message))
            response.headers['Content-Type'] = 'text/html'
            return response, status
        else:
            return "<pre>Unsupported Mimetype: {}</pre>".format(mimetype), 400

    except Exception as e:
        current_app.logger.error(e)
        current_app.logger.error(traceback.format_exc())
        return "<pre>" + traceback.format_exc() + "</pre>", 400


@git.route("/revert", defaults={'branch_or_ref': None}, methods=['GET', 'POST'])
@git.route("/revert/<path:branch_or_ref>", methods=['GET', 'POST'])
def revert(branch_or_ref):
    """Revert a commit.

    Returns:
    HTTP Response 201: If revert was possible and a revert commit was created
    HTTP Response 400: If revert did fail
    """
    try:
        current_app.config['quit'].repository.revert()
        current_app.config['quit'].syncAll()
        return '', 201
    except Exception as e:
        current_app.logger.error(e)
        current_app.logger.error(traceback.format_exc())
        return "<pre>" + traceback.format_exc() + "</pre>", 400
